
RELEASE_VERSION='0.3.571259835'
BUILD_DATE='Thu Jun 23 12:36:17 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-opyxurxq-project-28359933-concurrent-02cqf4 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2630958610'
GIT_COMMIT_SHA='e0ddb7656de46c4b3695307013f204b691cd3d35'
GIT_BRANCH='main'
